/*  RAxML-HPC, a program for sequential and parallel estimation of phylogenetic trees 
 *  Copyright March 2006 by Alexandros Stamatakis
 *
 *  Partially derived from
 *  fastDNAml, a program for estimation of phylogenetic trees from sequences by Gary J. Olsen
 *  
 *  and 
 *
 *  Programs of the PHYLIP package by Joe Felsenstein.
 *
 *  This program is free software; you may redistribute it and/or modify its
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 * 
 *
 *  For any other enquiries send an Email to Alexandros Stamatakis
 *  stamatak@ics.forth.gr
 *
 *  When publishing work that is based on the results from RAxML-HPC please cite:
 *  
 *  Alexandros Stamatakis: "An Efficient Program for phylogenetic Inference Using Simulated Annealing". 
 *  Proceedings of IPDPS2005,  Denver, Colorado, April 2005.
 *  
 *  AND
 *
 *  Alexandros Stamatakis, Thomas Ludwig, and Harald Meier: "RAxML-III: A Fast Program for Maximum Likelihood-based Inference of Large Phylogenetic Trees". 
 *  In Bioinformatics 21(4):456-463.
 */

#include <math.h>
#include <sys/mman.h>
#include <unistd.h>
#include <time.h> 
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "axml.h"
#ifdef PARALLEL
#include <mpi.h>
#endif

#define COMPUTE_TREE 0
#define TREE         1
#define FINALIZE     2
#define JOB_REQUEST  3

extern int optimizeTTinvocations;
extern int optimizeRatesInvocations;
extern int optimizeRateCategoryInvocations;
extern int optimizeAlphaInvocations;
extern int checkPointCounter;
extern int Thorough;
extern int treeStringLength;
extern double masterTime;
extern FILE   *INFILE, *permutationFile, *logFile, *infoFile;
extern char seq_file[2048];
extern char permFileName[1024], resultFileName[1024], 
  logFileName[1024], checkpointFileName[1024], infoFileName[1024], run_id[40], workdir[2048]; 
extern bestlist *bt;

extern int myId, numOfWorkers;

void printMultipleTree(tree *tr, int quoted, char *fileName)
{
  char *str = (char *)malloc(treeStringLength);
  char *outString;
  FILE *outf = fopen(fileName, "w");
  nodeptr startp;
  startp = tr->start;

  tr->start = tr->nodep[1];
  
  treeString(str, tr, tr->start->back, 1);  
  tr->start = startp; 
  outString = str;
  while(*outString++ != '(');
  fprintf(outf, "("); 
  if(!quoted)
  {
    while(*outString != '\0')
      {
	if(*outString != '\'')
	  fprintf(outf, "%c", *outString); 
	outString++;
      }
  }
      
  free(str);
  fclose(outf);  
}



void printBootstrapParallel(int quoted, char *fileName, char *str)
{
  char *outString;
  FILE *outf = fopen(fileName, "a");
 
  outString = str;
  while(*outString++ != '(');
  fprintf(outf, "("); 
  if(!quoted)
  {
    while(*outString != '\0')
      {
	if(*outString != '\'')
	  fprintf(outf, "%c", *outString); 
	outString++;
      }
  }
       
  fclose(outf);  
}


void printMultipleParallel(int quoted, char *fileName, char *str)
{
  char *outString;
  FILE *outf = fopen(fileName, "a");
 
  outString = str;
  while(*outString++ != '(');
  fprintf(outf, "("); 
  if(!quoted)
  {
    while(*outString != '\0')
      {
	if(*outString != '\'')
	  fprintf(outf, "%c", *outString); 
	outString++;
      }
  }
       
  fclose(outf);  
}






void printLogMultiple(tree *tr, analdef *adef, char *checkName, char *logName)
{
   
  if(adef->checkpoints)
    {
      char thisCheckPointName[2048], str[16];
      FILE *cpf, *lf;            
      
      lf = fopen(logName, "a");
      fprintf(lf,"%f %f %d\n", gettime() - masterTime, tr->likelihood, checkPointCounter); 
      fclose(lf);    
      
      strcpy(thisCheckPointName, checkName);
      sprintf(str, "%d", checkPointCounter);
      strcat(thisCheckPointName, str);      
      cpf = fopen(thisCheckPointName, "w");
      printTree(cpf, tr, 0);     
      fclose(cpf);    
      checkPointCounter++;
    }
  else
    {           
      FILE *lf;  
      lf = fopen(logName, "a");
      fprintf(lf,"%f %f\n", gettime() - masterTime, tr->likelihood); 
      fclose(lf);
    }    
}
    
    
  



void computeBIGRAPIDMULT (tree *tr, analdef *adef, int printLogFile, char *checkName, char *logName) 
{ 
  int i, rearrangementsMax, rearrangementsMin, noImproveCount, impr, goodTrav, bestTrav, which;                
  double lh, previousLh, difference, epsilon, currentLH, bestLH;              
  bestlist *bestT;
  
  checkPointCounter = 0;

  bestT = (bestlist *) malloc(sizeof(bestlist));
  bestT->ninit = 0;
  initBestTree(bestT, 1, tr->mxtips, tr->cdta->endsite);
    
  bt->ninit = 0;
  initBestTree(bt, 20, tr->mxtips, tr->cdta->endsite);  


  initInfoList(50);
 
  difference = 10.0;
  epsilon = 0.01;    
    
  Thorough = 0;           

  optimizeTTinvocations = 1;
  optimizeRatesInvocations = 1;
  optimizeRateCategoryInvocations = 1;
  optimizeAlphaInvocations = 1;
   
  
  optimizeModel(tr, adef, 1);   
  treeEvaluate(tr, 2);
  if(printLogFile)
    printLogMultiple(tr, adef, checkName, logName);

  saveBestTree(bestT, tr);   	  		      
  
  if(!adef->initialSet)
    bestTrav = determineRearrangementSetting(tr, adef, bestT);       
  else
    bestTrav = adef->initial;

  /*printf("bestTrav: %d\n", bestTrav);*/

  saveBestTree(bestT, tr);  
  impr = 1;

 
  while(impr)
    {     
      recallBestTree(bestT, 1, tr);	  
      optimizeModel(tr, adef, 1);    	 
      treeEvaluate(tr, 2);	 	 
      if(printLogFile)
	printLogMultiple(tr, adef, checkName, logName);
      saveBestTree(bestT, tr);	 
   
      lh = previousLh = tr->likelihood;
      
      treeOptimizeRapid(tr, 1, bestTrav, adef);
      
      impr = 0;
      
      for(i = 1; i <= bt->nvalid; i++)
	{  
	  recallBestTree(bt, i, tr);	    
	  treeEvaluate(tr, 0.25);  	 	
	  
	  difference = ((tr->likelihood > previousLh)? 
			tr->likelihood - previousLh: 
			previousLh - tr->likelihood); 	    
	  if(tr->likelihood > lh && difference > epsilon)
	    {
	      impr = 1;	       
	      lh = tr->likelihood;	       
	      saveBestTree(bestT, tr);		
	    }	   	   
	}	
    }
  
  Thorough = 1;
  impr = 1;
  
  while(1)
    {		
      recallBestTree(bestT, 1, tr);	
          	     	        
      if(impr)
	{	    	 
	  rearrangementsMin = 1;
	  rearrangementsMax = adef->stepwidth;	    
	}			  			
      else
	{		       	   
	  rearrangementsMax += adef->stepwidth;
	  rearrangementsMin += adef->stepwidth; 	        	      
	  if(rearrangementsMax > adef->max_rearrange)	     	     	 
	    goto cleanup; 	   
	}
      
      optimizeModel(tr, adef, 1);       
      treeEvaluate(tr, 2);					
      if(printLogFile)
	printLogMultiple(tr, adef, checkName, logName);
      previousLh = lh = tr->likelihood;	
      saveBestTree(bestT, tr);	
      
      treeOptimizeRapid(tr, rearrangementsMin, rearrangementsMax, adef);
	
      impr = 0;			      	
		
      for(i = 1; i <= bt->nvalid; i++)
	{	    		  	   
	  recallBestTree(bt, i, tr);	    
	  treeEvaluate(tr, 0.25);	    
	  
	  difference = ((tr->likelihood > previousLh)? 
			tr->likelihood - previousLh: 
			previousLh - tr->likelihood); 	    
	  if(tr->likelihood > lh && difference > epsilon)
	    {
	      impr = 1;	       
	      lh = tr->likelihood;	       
	      saveBestTree(bestT, tr);		
	    }	   	   
	}	
    }
    
 cleanup:   
  freeBestTree(bestT);
  free(bestT);
  freeBestTree(bt);
  freeInfoList();
}








boolean regetdata (analdef *adef, rawdata *rdta, tree *tr)
{
  int   i, j, k, l, basesread, basesnew, ch, my_i;
  int   meaning[256];          /*  meaning of input characters */ 
  char *nameptr;
  boolean  allread, firstpass;
  char buffer[300]; 
  int len;
  getnums (rdta);

  for (i = 0; i <= 255; i++) 
    meaning[i] = -1;

  if(adef->model == M_PROTEINS || adef->model == M_PROTEINSCAT)
    {          
      meaning['A'] =  0;  /* alanine */
      meaning['R'] =  1;  /* arginine */
      meaning['N'] =  2;  /*  asparagine*/
      meaning['D'] =  3;  /* aspartic */
      meaning['C'] =  4;  /* cysteine */
      meaning['Q'] =  5;  /* glutamine */
      meaning['E'] =  6;  /* glutamic */
      meaning['G'] =  7;  /* glycine */
      meaning['H'] =  8;  /* histidine */
      meaning['I'] =  9;  /* isoleucine */
      meaning['L'] =  10; /* leucine */
      meaning['K'] =  11; /* lysine */
      meaning['M'] =  12; /* methionine */
      meaning['F'] =  13; /* phenylalanine */
      meaning['P'] =  14; /* proline */
      meaning['S'] =  15; /* serine */
      meaning['T'] =  16; /* threonine */
      meaning['W'] =  17; /* tryptophan */
      meaning['Y'] =  18; /* tyrosine */
      meaning['V'] =  19; /* valine */
      meaning['B'] =  20; /* asparagine, aspartic 2 and 3*/
      meaning['Z'] =  21; /* glutamine glutamic 5 and 6*/
      meaning['X'] =  meaning['?'] = meaning['*'] = meaning['-'] = 22; /* all = 1.0 */
    }
  else
    {      
      meaning['A'] =  1;
      meaning['B'] = 14;
      meaning['C'] =  2;
      meaning['D'] = 13;
      meaning['G'] =  4;
      meaning['H'] = 11;
      meaning['K'] = 12;
      meaning['M'] =  3;
      meaning['N'] = 15;
      meaning['O'] = 15;
      meaning['R'] =  5;
      meaning['S'] =  6;
      meaning['T'] =  8;
      meaning['U'] =  8;
      meaning['V'] =  7;
      meaning['W'] =  9;
      meaning['X'] = 15;
      meaning['Y'] = 10;
      meaning['?'] = 15;
      meaning['-'] = 15;	
    }
  
  basesread = basesnew = 0;

  allread = FALSE;
  firstpass = TRUE;
  ch = ' ';

  while (! allread) 
    {
      for (i = 1; i <= tr->mxtips; i++) 
	{   	  
	  if (firstpass) 
	    {                      	       
	      ch = getc(INFILE);
	      while(ch == ' ' || ch == '\n' || ch == '\t')
		{
		  ch = getc(INFILE);
		}

	      my_i = 0;

	      do 
		{
		  buffer[my_i] = ch;
		  ch = getc(INFILE);		   
		  my_i++;
		  if(my_i >= nmlngth)
		    {
		      printf("ERROR Names are to long, adapt constant nmlngth in\n");
		      printf("axml.h AND dnapars.c, current setting %d\n", nmlngth);
		      exit(-1);
		    }
		}
	      while(ch !=  ' ' && ch != '\n' && ch != '\t');			     	
	    }


	  j = basesread;
	  while ((j < rdta->sites) && ((ch = getc(INFILE)) != EOF) && ((! adef->interleaved) || (ch != '\n'))) 
	    {
	      uppercase(& ch);
	      if ((meaning[ch] != -1) || ch == '.') 
		{
		  j++;
		  if (ch == '.') 
		    {
		      if (i != 1) ch = rdta->y[1][j];
		      else 
			{
			  printf("ERROR: Dot (.) found at site %d of sequence 1\n", j);
			  return  FALSE;
			}
		    }
		  rdta->y[i][j] = ch;
		}
	      else 
		if(whitechar(ch) || digitchar(ch)) ;
		else 
		  {
		    printf("ERROR: Bad base (%c) at site %d of sequence %d\n",
			   ch, j, i);
		    return  FALSE;
		  }
	    }

	    if (ch == EOF) 
	      {
		printf("ERROR: End-of-file at site %d of sequence %d\n", j, i);
		return  FALSE;
	      }

	    if (! firstpass && (j == basesread)) i--; 
	    else 
	      if (i == 1) basesnew = j;
	      else 
		if (j != basesnew) 
		  {
		    printf("ERROR: Sequences out of alignment\n");		    
		    printf("%d (instead of %d) residues read in sequence %d %s\n",
			   j - basesread, basesnew - basesread, i, tr->nameList[i]);
		    return  FALSE;
		  }
	    while (ch != '\n' && ch != EOF) ch = getc(INFILE);  /* flush line */
	}                                                  /* next sequence */
      firstpass = FALSE;
      basesread = basesnew;
      allread = (basesread >= rdta->sites);
    }

 
    
  for (j = 1; j <= tr->mxtips; j++)
    {
      for (i = 1; i <= rdta->sites; i++) 
	{
	  rdta->y[j][i] = meaning[rdta->y[j][i]];		
	}
    }   	

    return  TRUE;
  } /* getdata */


void  reMakeValuesAndWeigths(tree *tr, analdef *adef)
{
  int i, j, k;
  char buf;
  double  temp, wtemp;

  if(!adef->useWeightFile)
    {
      for (i = 1; i <= tr->rdta->sites; i++) 
	tr->rdta->wgt[i] = 1;
     
      tr->cdta->wgtsum = tr->rdta->sites;   
    }
  else
    {	
      inputweights(adef, tr->rdta, tr->cdta);
    }

  INFILE = fopen(seq_file, "r");
  regetdata(adef, tr->rdta, tr);
  fclose(INFILE); 
  
  makeweights(adef, tr->rdta, tr->cdta);
  makevalues(tr->rdta, tr->cdta);
  for (j = 0; j < tr->cdta->endsite; j++) 
    {
      tr->cdta->patrat[j] = temp = 1.0;
      tr->cdta->patratStored[j] = 1.0;
      tr->cdta->rateCategory[j] = 0;
      tr->cdta->wr[j]  = wtemp = temp * tr->cdta->aliaswgt[j];
      tr->cdta->wr2[j] = temp * wtemp;
    }
}




void doBootstrap(tree *tr, analdef *adef, rawdata *rdta, cruncheddata *cdta, int argc, char *argv[])
{
  int i, k, n, dummy;
  double  temp, wtemp, loopTime;
  MPI_Status msgStatus; 
  double bootLikelihood;

  n = adef->multipleRuns;

#ifdef DEBUG
  printf("Process %d in parallel Bootstrap\n", myId);
#endif

  strcpy(resultFileName, workdir); 
  strcat(resultFileName, "RAxML_bootstrap.");   
  strcat(resultFileName, run_id);  

  strcpy(infoFileName, workdir);
  strcat(infoFileName, "RAxML_info.");    
  strcat(infoFileName, run_id);      
             
  if(filexists(infoFileName))
    {
      printf("RAxML output files with the run ID <%s> already exist \n", run_id);
      printf("in directory %s ...... exiting\n", workdir);
      MPI_Finalize(); 
      exit(-1);
    }    

  MPI_Barrier(MPI_COMM_WORLD);

  makeweights(adef, rdta, cdta);
  makevalues(rdta, cdta);

  linkdata2tree(rdta, cdta, tr);    
  initModel(tr, rdta, cdta, adef); 

  if(myId == 0)    
    printModelAndProgramInfo(tr, adef, argc, argv);    
    
  treeStringLength = (tr->mxtips * (nmlngth+32)) + 256 + (tr->mxtips * 2);                         
  bt = (bestlist *) malloc(sizeof(bestlist));      
         
  if(myId == 0)
    {      
      int jobsSent = 0;
      int jobsReceived = n;
      
      while(jobsReceived > 0)
	{
	  char *str;
	 	  
	  str = (char *)malloc(sizeof(char) * treeStringLength);

	  MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &msgStatus); 
	  switch(msgStatus.MPI_TAG)
	    {
	    case JOB_REQUEST:
#ifdef DEBUG
	      printf("Master receiving work request from worker %d\n",  msgStatus.MPI_SOURCE);
#endif	      
	       MPI_Recv(&dummy, 1, MPI_INT, msgStatus.MPI_SOURCE, JOB_REQUEST, MPI_COMM_WORLD, &msgStatus);
	       if(jobsSent < n)
		 {
		   MPI_Send(&jobsSent, 1, MPI_INT, msgStatus.MPI_SOURCE, COMPUTE_TREE, MPI_COMM_WORLD);
#ifdef DEBUG
		   printf("Master snding job %d to worker %d\n",  jobsSent, msgStatus.MPI_SOURCE);
#endif		   
		   jobsSent++;
		 }
	       break;
	    case TREE:
	      MPI_Recv(str, treeStringLength, MPI_CHAR, msgStatus.MPI_SOURCE, TREE, MPI_COMM_WORLD, &msgStatus);

	      bootLikelihood = str_readTreeLikelihood(str);
#ifdef DEBUG
	      printf("--------> Master receiving tree from worker %d likelihood %f\n",  msgStatus.MPI_SOURCE, str_readTreeLikelihood(str));
#endif	 
	 
	      printBootstrapParallel(0, resultFileName, str);
	      printf("Bootstrap[%d] completed, bootstrap likelihood %f\n", n - jobsReceived, bootLikelihood);	 
	     
	      infoFile = fopen(infoFileName, "a");
	      fprintf(infoFile, "Bootstrap[%d] completed, bootstrap likelihood %f\n",  n - jobsReceived, bootLikelihood);     
	      fclose(infoFile);	    

	      jobsReceived--;

	      if(jobsSent < n)
		{
		  MPI_Send(&jobsSent, 1, MPI_INT, msgStatus.MPI_SOURCE, COMPUTE_TREE, MPI_COMM_WORLD);
#ifdef DEBUG
		  printf("Master snding job %d to worker %d\n",  jobsSent, msgStatus.MPI_SOURCE);
#endif	
		  jobsSent++;
		}
	      break;
	    }

	  free(str);
	}


      printf("\nOverall Time: %f\n", gettime() - masterTime);
      infoFile = fopen(infoFileName, "a");
      fprintf(infoFile, "\nOverall Time: %f\n", gettime() - masterTime);            
      fclose(infoFile);

      printf("All bootstrapped trees written to:      %s\n"    , resultFileName);   	          
      printf("Execution information file written to:  %s\n", infoFileName);

      for(i = 1; i < numOfWorkers; i++)
	{
	  MPI_Send(&dummy, 1, MPI_INT, i, FINALIZE, MPI_COMM_WORLD);
#ifdef DEBUG
	      printf("Master sending FINALIZE to worker %d\n",  i);
#endif	
	}
      MPI_Finalize();                 
      exit(1);
    }
  else  
    { 
      
      MPI_Send(&dummy, 1, MPI_INT, 0, JOB_REQUEST, MPI_COMM_WORLD);
#ifdef DEBUG
      printf("Worker %d sending job request to master\n",  myId);
#endif	

      while(1)
	{
	  char *str;
	  MPI_Probe(0, MPI_ANY_TAG, MPI_COMM_WORLD, &msgStatus); 
	  
	  str = (char *)malloc(sizeof(char) * treeStringLength);

	  switch(msgStatus.MPI_TAG)
	    {
	    case COMPUTE_TREE: 
	      MPI_Recv(&dummy, 1, MPI_INT, 0, COMPUTE_TREE, MPI_COMM_WORLD, &msgStatus);
#ifdef DEBUG
	      printf("Worker %d receiving job from master\n",  myId);
#endif	
	      optimizeRatesInvocations = 1;  
	      optimizeTTinvocations = 1;  
	      optimizeRateCategoryInvocations = 1;  
	      optimizeAlphaInvocations = 1;           
	      
	      initGlobalTip(tr, adef);              
	      
	      makeParsimonyTree(tr, adef);                       
	  
	      tr->likelihood = unlikely;

	      allocNodex(tr, adef); 	
	      initrav(tr, tr->start);
	      initrav(tr, tr->start->back);       
	      treeEvaluate(tr, 1);          
	  
	      computeBIGRAPIDMULT(tr, adef, 0, checkpointFileName, logFileName);     
	      
	      treeString(str, tr, tr->start->back, 1);
	      MPI_Send(str, treeStringLength, MPI_CHAR, 0, TREE, MPI_COMM_WORLD);
#ifdef DEBUG
      printf("Worker %d sending tree to master\n",  myId);
#endif	  

	      freeNodex(tr);            	  
	  
	      for(k = 1; k <= rdta->sites; k++)
		{	  
		  rdta->wgt[k] = rdta->wgt2[k] = 1;	 	  
		  cdta->aliaswgt[k] = 0;
		}
	  
	      reMakeValuesAndWeigths(tr, adef);                
	  
	      linkdata2tree(rdta, cdta, tr);    
	      initModel(tr, rdta, cdta, adef); 
	      break;
	    case FINALIZE:
	      MPI_Recv(&i, 1, MPI_INT, 0, FINALIZE, MPI_COMM_WORLD, &msgStatus);	      
#ifdef DEBUG
	      printf("Worker %d receiving FINALIZE %d\n",  myId);
#endif	  
	      MPI_Finalize(); 
	      exit(1);
	      break;
	    }
	  free(str);
	}
    }    
}


void packINT(int *msg, int ID, char *treeString, int length, double alpha)
{
  int i;

  msg[0] = ID;
  msg[1] = (int)(alpha * 10000000.0);

  for(i = 0; i < length; i++)
    {
      msg[i + 2] = (int)treeString[i];
    }
}

int unpackINT(int *msg, char *treeString, int length, double *alpha)
{
  int i;

  for(i = 0; i < length; i++)
    {
      treeString[i] = (char)msg[i+2];
    }

  *alpha = (((double)msg[1]) / 10000000.0);

  return msg[0];
}


void doMultiple(tree *tr, analdef *adef, rawdata *rdta, cruncheddata *cdta, int argc, char *argv[])
{
  int i, j, n, bestI, dummy;
  double  temp, wtemp, receivedLikelihood;;
  double loopTime, avgTime, bestLH, avgLH;
  double *times, *likelihoods;
  int jobID;
  char multID[1024];
  char buf[128]; 
  char temporaryFileName[2048];
  MPI_Status msgStatus;  
  char *msgStr;
  int  *msgInt; 
  int msgLength;
  double alpha;
  
  n = adef->multipleRuns;
 
  likelihoods = (double *)malloc(sizeof(double) * n);

  strcpy(permFileName, workdir);    
  strcpy(resultFileName, workdir); 
  strcpy(infoFileName, workdir);    
  strcat(permFileName, "RAxML_parsimonyTree.");
  strcat(resultFileName, "RAxML_result.");
  
  strcat(infoFileName, "RAxML_info.");    
  strcat(infoFileName, run_id);
  strcat(permFileName, run_id);
  strcat(resultFileName, run_id);
       
  if(filexists(infoFileName))
    {
      printf("RAxML output files with the run ID <%s> already exist \n", run_id);
      printf("in directory %s ...... exiting\n", workdir);
      MPI_Finalize(); 
      exit(-1);
    }

  MPI_Barrier(MPI_COMM_WORLD);
      
  makeweights(adef, rdta, cdta);
  makevalues(rdta, cdta);
  linkdata2tree(rdta, cdta, tr);    
  initModel(tr, rdta, cdta, adef);
        
  if(myId == 0)    
    printModelAndProgramInfo(tr, adef, argc, argv);    

  treeStringLength = (tr->mxtips * (nmlngth+32)) + 256 + (tr->mxtips * 2);                         
  bt = (bestlist *) malloc(sizeof(bestlist));            
  
  msgStr = (char *) malloc(sizeof(char) * treeStringLength);
  msgLength = treeStringLength + 1;
  msgInt = (int *)  malloc(sizeof(int)  * (msgLength));

  if(myId == 0)
    {      
      int jobsSent = 0;
      int jobsReceived = n;
      
      while(jobsReceived > 0)
	{       	 	        
	  MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &msgStatus); 
	  switch(msgStatus.MPI_TAG)
	    {
	    case JOB_REQUEST:
	       MPI_Recv(&dummy, 1, MPI_INT, msgStatus.MPI_SOURCE, JOB_REQUEST, MPI_COMM_WORLD, &msgStatus);
	       if(jobsSent < n)
		 {
		   MPI_Send(&jobsSent, 1, MPI_INT, msgStatus.MPI_SOURCE, COMPUTE_TREE, MPI_COMM_WORLD);
		   jobsSent++;
		 }
	       break;
	    case TREE:
	      MPI_Recv(msgInt, msgLength, MPI_INT, msgStatus.MPI_SOURCE, TREE, MPI_COMM_WORLD, &msgStatus);	
	      jobID = unpackINT(msgInt, msgStr, treeStringLength, &alpha);
#ifdef DEBUG
	      printf("MASTER receiving tree %s\n", msgStr);
#endif	      
	      receivedLikelihood =  str_readTreeLikelihood(msgStr);	      
	      likelihoods[jobID] =  receivedLikelihood;

	      printf("RUN[%d]: GTR+GAMMA Likelihood: %f alpha shape parameter %f\n", jobID,  receivedLikelihood, alpha);
	      infoFile = fopen(infoFileName, "a");
	      fprintf(infoFile, "RUN[%d]: GTR+GAMMA Likelihood: %f alpha shape parameter %f\n", jobID,  receivedLikelihood, alpha);
	      fclose(infoFile);

	      strcpy(multID, ".RUN.");     
	      sprintf(buf, "%d",  jobID);
	      strcat(multID, buf);    
	      strcpy(temporaryFileName,  resultFileName);
	      strcat(temporaryFileName,  multID); 
#ifdef DEBUG
	      printf("Result file name %s\n", temporaryFileName);
#endif
	      printMultipleParallel(0, temporaryFileName, msgStr);
	      jobsReceived--;

	      if(jobsSent < n)
		{
		  MPI_Send(&jobsSent, 1, MPI_INT, msgStatus.MPI_SOURCE, COMPUTE_TREE, MPI_COMM_WORLD);
		  jobsSent++;
		}
	      break;
	    }	
	}

      {
	avgTime = 0;
	avgLH = 0;
	bestLH = unlikely;
	bestI  = 0;

	for(i = 0; i < n; i++)
	  {	 
	    avgLH   += likelihoods[i];
	    if(likelihoods[i] > bestLH)
	      {
		bestLH = likelihoods[i];
		bestI  = i;
	      }
	  }

	printf("\nOverall Time: %f\n", gettime() - masterTime);
	printf("Average Likelihood: %f\n", (avgLH / n));
	printf("\n");
	printf("Best Likelihood in run number %d: likelihood %f\n", bestI, bestLH);

	infoFile = fopen(infoFileName, "a");
	fprintf(infoFile, "\nOverall Time: %f\n", gettime() - masterTime);
	fprintf(infoFile, "Average Likelihood: %f\n", (avgLH / n));
	fprintf(infoFile, "\n");
	fprintf(infoFile, "Best Likelihood in run number %d: likelihood %f\n", bestI, bestLH);
	fclose(infoFile);
      }

      if(adef->checkpoints)   
	printf("Checkpoints written to:                 %sRAxML_checkpoint.%s.RUN.%d to %d\n", workdir, run_id, 0, n-1);  
   
      printf("Parsimony starting trees written to:    %s.RUN.%d to %d\n", permFileName, 0, n-1);   	            
      printf("Final trees written to:                 %s.RUN.%d to %d\n", resultFileName,  0, n-1);  
      printf("Execution Log Files written to:         %sRAxML_log.%s.RUN.%d to %d\n", workdir, run_id, 0, n-1);   
      printf("Execution information file written to:  %s\n", infoFileName);


      for(i = 1; i < numOfWorkers; i++)
	MPI_Send(&dummy, 1, MPI_INT, i, FINALIZE, MPI_COMM_WORLD);
      MPI_Finalize(); 
      exit(1);
    }
  else
    {    
      MPI_Send(&dummy, 1, MPI_INT, 0, JOB_REQUEST, MPI_COMM_WORLD);

#ifdef DEBUG
      printf("Worker %d sending job request to master\n",  myId);
#endif	
      while(1)
	{

	  MPI_Probe(0, MPI_ANY_TAG, MPI_COMM_WORLD, &msgStatus); 	 

	  switch(msgStatus.MPI_TAG)
	    {
	    case COMPUTE_TREE: 
	      MPI_Recv(&jobID, 1, MPI_INT, 0, COMPUTE_TREE, MPI_COMM_WORLD, &msgStatus);
#ifdef DEBUG
	      printf("Worker %d receiving job from master\n",  myId);
#endif	
	      loopTime = gettime();
	      strcpy(multID, ".RUN.");     
	      sprintf(buf, "%d", jobID);
	      strcat(multID, buf);      
 
	      strcpy(checkpointFileName, workdir);
	      strcat(checkpointFileName, "RAxML_checkpoint.");
	      strcat(checkpointFileName, run_id);
	      strcat(checkpointFileName, multID);
	      strcat(checkpointFileName, ".");
      
	      strcpy(logFileName, workdir);  
	      strcat(logFileName, "RAxML_log.");
	      strcat(logFileName, run_id);
	      strcat(logFileName, multID);
                                            
	      optimizeRatesInvocations = 1;  
	      optimizeTTinvocations = 1;  
	      optimizeRateCategoryInvocations = 1;  
	      optimizeAlphaInvocations = 1;           

	      tr->likelihood = unlikely;

	      initGlobalTip(tr, adef);              	
	      makeParsimonyTree(tr, adef);        
	
	      strcpy(temporaryFileName,  permFileName);
	      strcat(temporaryFileName,  multID);   
	      printMultipleTree(tr, 0, temporaryFileName);
	      allocNodex(tr, adef); 	    
	      initrav(tr, tr->start);
	      initrav(tr, tr->start->back);          
	      treeEvaluate(tr, 1);            
	      computeBIGRAPIDMULT(tr, adef, 1, checkpointFileName, logFileName);       
      
	      resetBestTree(bt);

	      if(adef->model == M_GTRGAMMA && adef->mode != TREE_EVALUATION)
		{
		  modOpt(tr, adef);
		  loopTime = gettime() - loopTime;
#ifdef DEBUG		 
		  printf("Worker %d: Time %f GAMMA Likelihood: %f\n", myId, loopTime, tr->likelihood);
#endif		  
		  treeString(msgStr, tr, tr->start->back, 1);
#ifdef DEBUG
		  printf("Worker %d sending tree to master %s\n",  myId, msgStr);
#endif
		  packINT(msgInt, jobID, msgStr, treeStringLength, tr->alpha);

		  MPI_Send(msgInt, msgLength, MPI_INT, 0, TREE, MPI_COMM_WORLD);
		}   

	      freeNodex(tr);            
                       
	      if(adef->model == M_GTRCAT || adef->model == M_GTRMIX) 
		{	 	  	 
		  adef->model = M_GTRGAMMA;
		  allocNodex(tr, adef); 
		  initModel2(tr, rdta, cdta, adef);
		  optimizeRatesInvocations = 1;  
		  optimizeTTinvocations = 1;  
		  optimizeRateCategoryInvocations = 1;  
		  optimizeAlphaInvocations = 1;
		  initGlobalTip(tr, adef);
		  
		  for(j = 0; j < tr->cdta->endsite; j++)
		    tr->cdta->wr[j] = tr->cdta->aliaswgt[j];
		  
		  /* initrav(tr, tr->start);
		  initrav(tr, tr->start->back);
		  treeEvaluate(tr, 1);*/
	  
		  modOpt(tr, adef);
		  
		  loopTime = gettime() - loopTime;		  
#ifdef DEBUG
		  printf("Worker %d, Time %f, GAMMA Likelihood: %f\n", myId, loopTime, tr->likelihood);
#endif		 	 
		  treeString(msgStr, tr, tr->start->back, 1);

#ifdef DEBUG
		  printf("Worker %d sending tree to master %s\n",  myId, msgStr);
#endif
		  packINT(msgInt, jobID, msgStr, treeStringLength, tr->alpha);

		  MPI_Send(msgInt, msgLength, MPI_INT, 0, TREE, MPI_COMM_WORLD);
		
		  freeNodex(tr);   
		  adef->model = M_GTRCAT;
		}              	     
        
	      for (j = 0; j < tr->cdta->endsite; j++) 
		{
		  tr->cdta->patrat[j] = temp = 1.0;
		  tr->cdta->patratStored[j] = 1.0;
		  tr->cdta->rateCategory[j] = 0;
		  tr->cdta->wr[j]  = wtemp = temp * tr->cdta->aliaswgt[j];
		  tr->cdta->wr2[j] = temp * wtemp;
		}    

	      initModel(tr, rdta, cdta, adef);
	      break;
	    case FINALIZE:
	      MPI_Recv(&i, 1, MPI_INT, 0, FINALIZE, MPI_COMM_WORLD, &msgStatus);
#ifdef DEBUG
	      printf("Worker %d receiving FINALIZE %d\n",  myId);
#endif	     
	      MPI_Finalize();
	      exit(1);
	      break;
	    }       
	}
    }  
}
